//
// "$Id: $"
//
// Fontconfig based font utilities for FLTK
//
// Copyright 2001-2014 Bill Spitzak and others.
//
// This library is free software. Distribution and use rights are outlined in
// the file "COPYING" which should have been included with this file.  If this
// file is missing or damaged, see the license at:
//
//     http://www.fltk.org/COPYING.php
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

#ifndef FL_DOXYGEN

#include <X11/Xft/Xft.h>

/* The following functions are to test how we might use fontconfig to
 * find a suitable font for a portion of text. */
#if defined (FLTK_ABI_VERSION) && (FLTK_ABI_VERSION >= 10303)
extern "C" {

int fl_xft_font_finder_for_UTF8_text(const char *txt,
                                     const int is_mono,
                                     const int is_bold,
                                     const int is_italic,
                                     const int pix_size)
{
  FcCharSet *text_chars = FcCharSetCreate();
  FcPattern *txt_pat = FcPatternCreate();

  // map the input UTF8 text into a charset
  const FcChar8 *pc = (const FcChar8 *)txt;
  int sz = strlen(txt);
  // just in case...
  if (sz > 2048) sz = 2048;

  while (sz > 0) { // loop until we have eaten all the UTF8 chars
    FcChar32 ucs32;
    // Discard ctrl chars etc... This also removes tabs and so forth
    // this is far from thorough though - there are many other ranges we
    // should elide at this point!
    while ((*pc < 0x20) && (sz > 0)) {
      ++pc;
      --sz;
    }

    int used = FcUtf8ToUcs4(pc, &ucs32, sz);
    if (used < 1) {
      fprintf(stderr, "Non UTF-8 input found\n");
      return -1;
    }
    else { // char was OK, advance to next char position
      sz -= used;
      pc += used;
      // add unicode char to our test charset
      FcBool res = FcCharSetAddChar(text_chars, ucs32);
      if (res == FcFalse) {
        fprintf(stderr, "FcCharSetAddChar Failure\n");
        return -1;
      } // bad cahr set add
    } // OK USC4 read from txt
  } // while string...

  // Construct a match pattern for the font we want...
  FcPatternAddDouble (txt_pat, FC_PIXEL_SIZE, pix_size);
  if (is_mono != 0) {
    FcPatternAddInteger(txt_pat, FC_SPACING, FC_MONO);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_SPACING, FC_PROPORTIONAL);
  }

  if (is_bold != 0) {
    FcPatternAddInteger(txt_pat, FC_WEIGHT, FC_WEIGHT_BOLD);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_WEIGHT, FC_WEIGHT_NORMAL);
  }

  if (is_italic != 0) {
    FcPatternAddInteger(txt_pat, FC_SLANT, FC_SLANT_ITALIC);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_SLANT, FC_SLANT_ROMAN);
  }
  // the set of characters we need to cover...
  FcPatternAddCharSet(txt_pat, FC_CHARSET, text_chars);

  FcConfigSubstitute(NULL, txt_pat, FcMatchFont); // all system fonts
  FcDefaultSubstitute(txt_pat);

  FcResult match_result;
  FcFontSet *fnt_set = NULL;
  fnt_set = FcFontSort(NULL, txt_pat, FcTrue, NULL, &match_result); // trim coverage
//  fnt_set = FcFontSort(NULL, txt_pat, FcFalse, NULL, &match_result); // do not trim coverage

  int num_fonts = (-1);
  if (fnt_set) {
    num_fonts = fnt_set->nfont;

    if (num_fonts > 64) num_fonts = 64;
    for (int i = 0; i < num_fonts; ++i) {
      FcPattern *nxt_pat = fnt_set->fonts[i];
      // How many of the required glyphs does this *not* cover?
      int missing = -1;
      FcValue fcv;
      FcPatternGet(nxt_pat, FC_CHARSET, 0, &fcv);
      if (fcv.type == FcTypeCharSet){
        const FcCharSet *fnt_charset = fcv.u.c;
        FcBool res = FcCharSetIsSubset(text_chars, fnt_charset);
        if (res != FcFalse) {
          // every glyph in text_chars is available in charset
          missing = 0;
        }
        else {
          // Some glyphs are missing - but how many?
          missing = (int)FcCharSetSubtractCount(text_chars, fnt_charset); // how many we DO NOT have
        } // count missing glyphs
      }

      char *picked_name =  (char *)FcNameUnparse(nxt_pat); // this returns a ":" separated list of pattern elements
      char *pc = strchr(picked_name, ':'); // find first element of name
      if (pc) *pc = 0;
      printf("Found font:%3d %s, is missing %d glyph%s\n", i, picked_name, missing, ((missing != 1)?"s":" "));
      free(picked_name);
    }
    FcFontSetDestroy(fnt_set);
  }

  if (txt_pat) FcPatternDestroy(txt_pat);
  if (text_chars) FcCharSetDestroy(text_chars);

  return num_fonts;
} // fl_xft_font_finder_for_UTF8_text

///////////////////////////////////////////////////////////////

int fl_xft_font_set_find (const char *txt,
                          const int is_mono,
                          const int is_bold,
                          const int is_italic,
                          const int pix_size,
                          char **names,
                          int *N)
{
  FcCharSet *text_chars = FcCharSetCreate();
  FcPattern *txt_pat = FcPatternCreate();
  const int max_names = *N;

  // map the input UTF8 text into a charset
  const FcChar8 *pc = (const FcChar8 *)txt;
  int sz = strlen(txt);
  // just in case...
  if (sz > 2048) sz = 2048;

  while (sz > 0) { // loop until we have eaten all the UTF8 chars
    FcChar32 ucs32;
    // Discard ctrl chars etc... This also removes tabs and so forth
    // this is far from thorough though - there are many other ranges we
    // should elide at this point!
    while ((*pc < 0x20) && (sz > 0)) {
      ++pc;
      --sz;
    }

    int used = FcUtf8ToUcs4(pc, &ucs32, sz);
    if (used < 1) {
      fprintf(stderr, "Non UTF-8 input found\n");
      return -1;
    }
    else { // char was OK, advance to next char position
      sz -= used;
      pc += used;
      // add unicode char to our test charset
      FcBool res = FcCharSetAddChar(text_chars, ucs32);
      if (res == FcFalse) {
        fprintf(stderr, "FcCharSetAddChar Failure\n");
        return -1;
      } // bad cahr set add
    } // OK USC4 read from txt
  } // while string...

  // Construct a match pattern for the font we want...
  FcPatternAddDouble (txt_pat, FC_PIXEL_SIZE, pix_size);
  if (is_mono != 0) {
    FcPatternAddInteger(txt_pat, FC_SPACING, FC_MONO);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_SPACING, FC_PROPORTIONAL);
  }

  if (is_bold != 0) {
    FcPatternAddInteger(txt_pat, FC_WEIGHT, FC_WEIGHT_BOLD);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_WEIGHT, FC_WEIGHT_NORMAL);
  }

  if (is_italic != 0) {
    FcPatternAddInteger(txt_pat, FC_SLANT, FC_SLANT_ITALIC);
  }
  else {
    FcPatternAddInteger(txt_pat, FC_SLANT, FC_SLANT_ROMAN);
  }
  // the set of characters we need to cover...
  FcPatternAddCharSet(txt_pat, FC_CHARSET, text_chars);

  FcConfigSubstitute(NULL, txt_pat, FcMatchFont); // all system fonts
  FcDefaultSubstitute(txt_pat);

  FcResult match_result;
  FcFontSet *fnt_set = NULL;
  fnt_set = FcFontSort(NULL, txt_pat, FcTrue, NULL, &match_result); // trim coverage
//  fnt_set = FcFontSort(NULL, txt_pat, FcFalse, NULL, &match_result); // do not trim coverage

  int missing = -1;
  int faces_picked = 0;
  FcCharSet *miss_set;

  while ((missing != 0) && (faces_picked < max_names)) {
    if (fnt_set) {
      const FcCharSet *fnt_charset;
      // assume the first pick is the "best" base font...
      FcPattern *nxt_pat = fnt_set->fonts[0];
      // How many of the required glyphs does this *not* cover?
      FcValue fcv;
      FcPatternGet(nxt_pat, FC_CHARSET, 0, &fcv);
      if (fcv.type == FcTypeCharSet) {
        fnt_charset = fcv.u.c;
        FcBool res = FcCharSetIsSubset(text_chars, fnt_charset);
        if (res != FcFalse) {
          // every glyph in text_chars is available in charset
          missing = 0;
        }
        else {
          // Some glyphs are missing - but how many?
          missing = (int)FcCharSetSubtractCount(text_chars, fnt_charset); // how many we DO NOT have
        } // count missing glyphs
      }
      else {
        break; // bad fcv so exit
      }

      char *picked_name =  (char *)FcNameUnparse(nxt_pat); // this returns a ":" separated list of pattern elements
      char *pc = strchr(picked_name, ':'); // find first element of name
      if (pc) *pc = 0;
      pc = strchr(picked_name, ','); // find first element of name
      if (pc) *pc = 0;
      printf("Chose font:%3d %s, still missing %d glyph%s\n", faces_picked, picked_name, missing, ((missing != 1)?"s":" "));

      int namelen = strlen(picked_name) + 2;
      names[faces_picked] = (char *)malloc(namelen + 1);
      snprintf(names[faces_picked], namelen, " %s", picked_name);
//      names[faces_picked] = strdup(picked_name);
      free(picked_name);
      ++faces_picked;

      if (missing != 0) {
        // make a charset for just the missing glyphs
        miss_set = FcCharSetSubtract (text_chars, fnt_charset);
        if (text_chars) FcCharSetDestroy(text_chars);
        if (miss_set) {
          text_chars = FcCharSetCopy(miss_set);
          FcCharSetDestroy(miss_set);
        }
        FcPatternDel(txt_pat, FC_CHARSET); // remove old charset
        FcPatternAddCharSet(txt_pat, FC_CHARSET, text_chars); // add new charset
        FcFontSetDestroy(fnt_set); // discard old fnt_set
        fnt_set = FcFontSort(NULL, txt_pat, FcTrue, NULL, &match_result); // get new set
      }
      else {
        FcFontSetDestroy(fnt_set);
      }
    } // valid fnt_set
    else {
      break; // bad fnt_set, exit now
    } // invalid fnt_set
  } // while missing glyphs...

  if (txt_pat) FcPatternDestroy(txt_pat);
  if (text_chars) FcCharSetDestroy(text_chars);

  *N = faces_picked;

  return missing;
} // fl_xft_font_set_find


///////////////////////////////////////////////////////////////
// Search for fonts that have similar properties to some face
// we have already loaded.
int fl_xft_font_set_find_similar (const char *txt,
                                  const Fl_Font base_fnt,
                                  char **names,
                                  int *N)
{
  FcCharSet *text_chars = FcCharSetCreate();
  //FcPattern *txt_pat = FcPatternCreate();
  const int max_names = *N;

  Fl_Fontsize base_sz = fl_size();
  Fl_Font orig_fnt = fl_font();
  char style_char = ' ';

  Fl_Fontdesc *tst_fnt = fl_fonts + base_fnt;
  Fl_Font_Descriptor* fnt_desc;
  // is the face loaded?
  if (tst_fnt->name) { // name is non-NULL, so font is set
    style_char = tst_fnt->name[0];
    fnt_desc = tst_fnt->first;
    if (!fnt_desc) { // font descriptor is not populated yet, set it now
      fl_font(base_fnt, base_sz);
      fnt_desc = tst_fnt->first;
      fl_font(orig_fnt, base_sz);
    }
    if (!fnt_desc) { // failed to populate this font description
      printf("Failed to resolve base font\n");
      return -1;
    }
  }
  else {
    printf("Base font not set\n");
    return -1;
  }

  XftFont* xft_fnt = fnt_desc->font;
  FcPattern *src_pattern = xft_fnt->pattern;
  // is the base font AA capable? If so, we want to get an AA substitute too...
  FcBool want_aa;
//  FcPatternGetBool(src_pattern, FC_ANTIALIAS, 0, &want_aa);
  {
    FcValue fcb;
    FcPatternGet(src_pattern, FC_ANTIALIAS, 0, &fcb);
    if (fcb.type == FcTypeBool) {
      want_aa = fcb.u.b;
    }
    else {
    // No FC_ANTIALIAS set in pattern
    // does this mean that it IS AA then? Assume not...
      want_aa = FcFalse;
    }
  }

  // map the input UTF8 text into a charset
  const FcChar8 *pc = (const FcChar8 *)txt;
  int sz = strlen(txt);
  // just in case...
  if (sz > 2048) sz = 2048;

  while (sz > 0) { // loop until we have eaten all the UTF8 chars
    FcChar32 ucs32;
    // Discard ctrl chars etc... This also removes tabs and so forth
    // this is far from thorough though - there are many other ranges we
    // should elide at this point!
    while ((*pc < 0x20) && (sz > 0)) {
      ++pc;
      --sz;
    }

    int used = FcUtf8ToUcs4(pc, &ucs32, sz);
    if (used < 1) {
      fprintf(stderr, "Non UTF-8 input found\n");
      return -1;
    }
    else { // char was OK, advance to next char position
      sz -= used;
      pc += used;
      // add unicode char to our test charset
      FcBool res = FcCharSetAddChar(text_chars, ucs32);
      if (res == FcFalse) {
        fprintf(stderr, "FcCharSetAddChar Failure\n");
        return -1;
      } // bad cahr set add
    } // OK USC4 read from txt
  } // while string...

  // Construct a match pattern for the font we want,
  // based on the font index we were passed...
  FcObjectSet *obj_set = FcObjectSetBuild(FC_STYLE, FC_SLANT, FC_WEIGHT,
                                          FC_SIZE, FC_ASPECT,
                                          FC_PIXEL_SIZE, FC_SPACING, NULL);
  FcPattern *txt_pat = FcPatternFilter(src_pattern, obj_set);

  // the set of characters we need to cover...
  FcPatternAddCharSet(txt_pat, FC_CHARSET, text_chars);

  if (want_aa != FcFalse) {
    FcPatternAddBool(txt_pat, FC_ANTIALIAS, FcTrue);
    FcPatternAddBool(txt_pat, FC_SCALABLE, FcTrue);
  }

  FcConfigSubstitute(NULL, txt_pat, FcMatchFont); // all system fonts
  FcDefaultSubstitute(txt_pat);

  FcResult match_result;
  FcFontSet *fnt_set = NULL;
  fnt_set = FcFontSort(NULL, txt_pat, FcTrue, NULL, &match_result); // trim coverage

  int missing = -1;
  int faces_picked = 0;
  FcCharSet *miss_set;

  while ((missing != 0) && (faces_picked < max_names)) {
    if (fnt_set) {
      const FcCharSet *fnt_charset;
      // assume the first pick is the "best" base font...
      FcPattern *nxt_pat = fnt_set->fonts[0];
      // How many of the required glyphs does this *not* cover?
      FcValue fcv;
      FcPatternGet(nxt_pat, FC_CHARSET, 0, &fcv);
      if (fcv.type == FcTypeCharSet) {
        fnt_charset = fcv.u.c;
        FcBool res = FcCharSetIsSubset(text_chars, fnt_charset);
        if (res != FcFalse) {
          // every glyph in text_chars is available in charset
          missing = 0;
        }
        else {
          // Some glyphs are missing - but how many?
          missing = (int)FcCharSetSubtractCount(text_chars, fnt_charset); // how many we DO NOT have
        } // count missing glyphs
      }
      else {
        break; // bad fcv so exit
      }

      char *picked_name =  (char *)FcNameUnparse(nxt_pat); // this returns a ":" separated list of pattern elements
      char *pc = strchr(picked_name, ':'); // find first element of name
      if (pc) *pc = 0;
      pc = strchr(picked_name, ','); // find first element of name
      if (pc) *pc = 0;
      printf("Chose font:%3d %s, still missing %d glyph%s\n", faces_picked, picked_name, missing, ((missing != 1)?"s":" "));

      int namelen = strlen(picked_name) + 2;
      names[faces_picked] = (char *)malloc(namelen + 1);
      snprintf(names[faces_picked], namelen, "%c%s", style_char, picked_name);
//      names[faces_picked] = strdup(picked_name);
      free(picked_name);
      ++faces_picked;

      if (missing != 0) {
        // make a charset for just the missing glyphs
        miss_set = FcCharSetSubtract (text_chars, fnt_charset);
        if (text_chars) FcCharSetDestroy(text_chars);
        if (miss_set) {
          text_chars = FcCharSetCopy(miss_set);
          FcCharSetDestroy(miss_set);
        }
        FcPatternDel(txt_pat, FC_CHARSET); // remove old charset
        FcPatternAddCharSet(txt_pat, FC_CHARSET, text_chars); // add new charset
        FcFontSetDestroy(fnt_set); // discard old fnt_set
        fnt_set = FcFontSort(NULL, txt_pat, FcTrue, NULL, &match_result); // get new set
      }
      else {
        FcFontSetDestroy(fnt_set);
      }
    } // valid fnt_set
    else {
      break; // bad fnt_set, exit now
    } // invalid fnt_set
  } // while missing glyphs...

  if (txt_pat) FcPatternDestroy(txt_pat);
  if (text_chars) FcCharSetDestroy(text_chars);
  if (obj_set) FcObjectSetDestroy(obj_set);

  *N = faces_picked;

  return missing;
} // fl_xft_font_set_find_similar


} /* extern C */
#endif /* FLTK_ABI_VERSION */
/* End of font finder functions... */

#endif /* not FL_DOXYGEN */

//
// End of "$Id: $"
//
