//
// "$Id: unittest_text.cxx 8864 2011-07-19 04:49:30Z greg.ercolano $"
//
// Unit tests for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2014 by Bill Spitzak and others.
//
// This library is free software. Distribution and use rights are outlined in
// the file "COPYING" which should have been included with this file.  If this
// file is missing or damaged, see the license at:
//
//     http://www.fltk.org/COPYING.php
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

#include <FL/Fl_Box.H>
#include <FL/fl_draw.H>

extern "C" {
  extern void set_xft_substitution_state (const int state);
  // Search for fonts that have similar properties to some face
  // we have already loaded.
  int fl_xft_font_set_find_similar (const char *txt,
                                    const Fl_Font base_fnt,
                                    char **names,
                                    int *N);
} /* extern C */

const char test_text[] =
"RU: \xd1\x80\xd1\x83\xcc\x81\xd1\x81\xd1\x81\xd0\xba\xd0\xb8\xd0\xb9, "
"JP: \xe6\x97\xa5\xe6\x9c\xac\xe5\x9b\xbd "
"MusicalG\xF0\x9D\x84\x9E  clef";

//
// --- fl_text_extents() tests -----------------------------------------------
//
class TextExtentsTest : public Fl_Widget
{
  void DrawTextAndBoxes(const char *txt, int X, int Y) {
    int wo = 0, ho = 0;
    int dx, dy;
    // First, we draw the bounding boxes (fl_measure and fl_text_extents)
    // draw fl_measure() typographical bounding box
    fl_measure(txt, wo, ho, 0);
    int desc = fl_descent();
    fl_color(FL_RED);
    fl_rect(X, Y-ho+desc, wo, ho);
    // draw fl_text_extents() glyph bounding box
    fl_text_extents(txt, dx, dy, wo, ho);
    fl_color(FL_GREEN);
    fl_rect(X+dx, Y+dy, wo, ho);
    // Then we draw the text to show how it fits inside each of the two boxes
    fl_color(FL_BLACK);
    fl_draw(txt, X, Y);
  } // DrawTextAndBoxes

public:
  static Fl_Widget *create() {
    set_xft_substitution_state(1);
    Fl::set_font(FL_FREE_FONT+1, " FreeSerif");
    Fl_Font nxt_idx = FL_FREE_FONT + 2;
    int max_names = 6;
    char *names[6];
//    int missing = fl_xft_font_set_find_similar (test_text, (FL_FREE_FONT+1), names, &max_names);
    int missing = fl_xft_font_set_find_similar (test_text, FL_HELVETICA, names, &max_names);

    printf("Selected %d fonts\n", max_names);
    for (int idx = 0; idx < max_names; ++idx) {
      printf("%d: %s \n", idx, names[idx]);
      Fl::set_font(nxt_idx, names[idx]);
      ++nxt_idx;
    }
    printf("%d glyphs are missing\n", missing);

    return new TextExtentsTest(TESTAREA_X, TESTAREA_Y, TESTAREA_W, TESTAREA_H);
  } // create

  // constructor
  TextExtentsTest(int x, int y, int w, int h) : Fl_Widget(x, y, w, h) {}

  void draw(void) {
    int x0 = x(); // origin is current window position for Fl_Box
    int y0 = y();
    int w0 = w();
    int h0 = h();
    fl_push_clip(x0, y0, w0, h0); // reset local clipping
    {
      // set the background colour - slightly off-white to enhance the green bounding box
      fl_color(fl_gray_ramp(FL_NUM_GRAY - 3));
      fl_rectf(x0, y0, w0, h0);

      fl_font(FL_HELVETICA, 30);
      int xx = x0+55;
      int yy = y0+40;
      DrawTextAndBoxes("!abcdeABCDE\"#A", xx, yy); yy += 50;	// mixed string
      DrawTextAndBoxes("oacs",     xx, yy); xx += 100;		// small glyphs
      DrawTextAndBoxes(test_text,   xx, yy); yy += 50; xx -= 100;	// glyphs with descenders
      DrawTextAndBoxes("````````", xx, yy); yy += 50;		// high small glyphs
      DrawTextAndBoxes("--------", xx, yy); yy += 50;		// mid small glyphs
      DrawTextAndBoxes("________", xx, yy); yy += 50;		// low small glyphs

      fl_font(FL_HELVETICA, 14);
      fl_color(FL_RED);  fl_draw("fl_measure bounding box in RED",       xx, yy); yy += 20;
      fl_color(FL_GREEN); fl_draw("fl_text_extents bounding box in GREEN", xx, yy);
      fl_color(FL_BLACK);
      xx = x0 + 10;  yy += 30;
      fl_draw("NOTE: On systems with text anti-aliasing (e.g. OSX Quartz)", xx, yy);
      w0 = h0 = 0; fl_measure("NOTE: ", w0, h0, 0);
      xx += w0; yy += h0;
      fl_draw("text may leak slightly outside the fl_text_extents()", xx, yy);
    }
    fl_pop_clip(); // remove the local clip
  } // draw
};

UnitTest textExtents("rendering text", TextExtentsTest::create);

//
// End of "$Id: unittest_text.cxx 8864 2011-07-19 04:49:30Z greg.ercolano $"
//
